package net.javainthebox.heapindicator;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.util.Collection;
import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Timer;
import org.openide.util.ImageUtilities;

import com.sun.tools.visualvm.application.Application;
import com.sun.tools.visualvm.charts.ChartFactory;
import com.sun.tools.visualvm.charts.SimpleXYChartDescriptor;
import com.sun.tools.visualvm.charts.SimpleXYChartSupport;
import com.sun.tools.visualvm.core.datasource.DataSource;
import com.sun.tools.visualvm.core.datasupport.DataRemovedListener;
import com.sun.tools.visualvm.core.options.GlobalPreferences;
import com.sun.tools.visualvm.core.ui.DataSourceView;
import com.sun.tools.visualvm.core.ui.components.DataViewComponent;
import com.sun.tools.visualvm.tools.jmx.JmxModel;
import com.sun.tools.visualvm.tools.jmx.JmxModelFactory;
import com.sun.tools.visualvm.tools.jmx.JvmMXBeans;
import com.sun.tools.visualvm.tools.jmx.JvmMXBeansFactory;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.UndeclaredThrowableException;

public class HeapIndicator extends DataSourceView {
    // アイコンのパス
    private static final String IMAGE_PATH 
            = "net/javainthebox/heapindicator/heap.png";

    // JMX をカプセル化する JmxMobel オブジェクト
    private JmxModel jmx;
    
    // メモリに関する MXBean
    private MemoryPoolMXBean edenMXBean;
    private MemoryPoolMXBean survivorMXBean;
    private MemoryPoolMXBean tenuredMXBean;

    // Eden 領域チャート
    private SimpleXYChartSupport edenChartSupport;
    // Survivor 領域チャート
    private SimpleXYChartSupport survivorChartSupport;
    // Tenured 領域チャート
    private SimpleXYChartSupport tenuredChartSupport;

    // データ更新用配列
    private long[] values = new long[2];
    private String[] details = new String[3];

    // 値更新用のタイマ
    private Timer timer;
    private int pollingTime;

    public HeapIndicator(Application application) {
        super(application, 
              "Heap",
              ImageUtilities.loadImage(IMAGE_PATH),
              100, false);

        // JMX を用いてアプリケーションとの接続を行なう
        jmx = JmxModelFactory.getJmxModelFor(application);
        // JMX の接続が切れたら、タイマーを停止させる
        jmx.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent event) {
                if ("connectionState".equals(event.getPropertyName())
                        && event.getNewValue() == JmxModel.ConnectionState.DISCONNECTED) {
                    timer.stop();
                }
             }
        });

        // プラットフォームMBeanServerに相当するJvmMXBeansオブジェクトを取得
        JvmMXBeans mxbeans = JvmMXBeansFactory.getJvmMXBeans(jmx);
        Collection<MemoryPoolMXBean> poolMXBeans = mxbeans.getMemoryPoolMXBeans();
        for (MemoryPoolMXBean poolMXBean: poolMXBeans) {
            String name = poolMXBean.getName();
            if (name.contains("Eden Space")) {
                edenMXBean = poolMXBean;
            } else if (name.contains("Survivor Space")) {
                survivorMXBean = poolMXBean;
            } else if (name.contains("Tenured Gen") || name.contains("Old Gen")) {
                tenuredMXBean = poolMXBean;
            }
        }
    }
    
    @Override
    protected DataViewComponent createComponent() {
        // マスタービューに表示するコンテンツ
        JPanel generalPanel = new JPanel();
        generalPanel.setOpaque(false);

        // マスタービュー
        DataViewComponent.MasterView masterView
                = new DataViewComponent.MasterView("Heap Indicator", null, generalPanel);

        DataViewComponent.MasterViewConfiguration masterConfiguration =
                new DataViewComponent.MasterViewConfiguration(false);

        // DataViewComponentオブジェクトの生成
        DataViewComponent dataViewComponent
                = new DataViewComponent(masterView, masterConfiguration);

        // Eden領域のチャート作成
        edenChartSupport = createChartSupport(dataViewComponent,
                                             "Eden",
                                              DataViewComponent.TOP_LEFT,
                                              edenMXBean);
        // Survivor領域のチャート作成
        survivorChartSupport = createChartSupport(dataViewComponent,
                                             "Survivor",
                                              DataViewComponent.TOP_RIGHT,
                                              survivorMXBean);
        // Tenured領域のチャート作成
        tenuredChartSupport = createChartSupport(dataViewComponent,
                                             "Tenured",
                                              DataViewComponent.BOTTOM_RIGHT,
                                              tenuredMXBean);
        
        // 詳細ビューの設定
        dataViewComponent.configureDetailsArea(
                new DataViewComponent.DetailsAreaConfiguration("Eden", true),
                DataViewComponent.TOP_LEFT);
        dataViewComponent.configureDetailsArea(
                new DataViewComponent.DetailsAreaConfiguration("Survivor", true),
                DataViewComponent.TOP_RIGHT);
        dataViewComponent.configureDetailsArea(
                new DataViewComponent.DetailsAreaConfiguration("Tenured", true),
                DataViewComponent.BOTTOM_RIGHT);

        // 周期的に値を更新する
        ActionListener task = new ActionListener () {
            @Override
            public void actionPerformed(ActionEvent event) {
                if (jmx.getConnectionState() == JmxModel.ConnectionState.CONNECTED) {
                    // JMX との通信が確立している場合
                    try {
                        updateValues();
                    } catch (UndeclaredThrowableException ex) {
                        // 情報取得中に通信が切れた場合
                    }
                }
            }
        };
        timer = new Timer(pollingTime * 1000, task);
        timer.setDelay(pollingTime * 1000);
        timer.start();

        // 終了時にタイマを停止させる
        DataSource dataSource = getDataSource();
        dataSource.notifyWhenRemoved(new DataRemovedListener() {
            @Override
            public void dataRemoved(Object arg0) {
                timer.stop();
            }
        });
        
        return dataViewComponent;
    }

    private void updateValues() {
        // Eden使用量の更新
        if (edenMXBean != null) {
            updateValue(edenMXBean, edenChartSupport);
        }

        // Survivor使用量の更新
        if (survivorMXBean != null) {
            updateValue(survivorMXBean, survivorChartSupport);
        }

        // Tenured使用量の更新
        if (tenuredMXBean != null) {
            updateValue(tenuredMXBean, tenuredChartSupport);
        }
    }

    private void updateValue(MemoryPoolMXBean mxBean, SimpleXYChartSupport chartSupport) {
        MemoryUsage usage = mxBean.getUsage();

        // チャートに値を追加
        values[0] = usage.getCommitted();
        values[1] = usage.getUsed();
        chartSupport.addValues(System.currentTimeMillis(), values);

        // 詳細情報を更新
        details[0] = chartSupport.formatBytes(usage.getMax());
        details[1] = chartSupport.formatBytes(usage.getCommitted());
        details[2] = chartSupport.formatBytes(usage.getUsed());
        chartSupport.updateDetails(details);
    }

    private SimpleXYChartSupport createChartSupport(DataViewComponent dataViewComponent,
                                                      String name,
                                                      int position,
                                                      MemoryPoolMXBean mxBean) {

        // メニューバーの[ツール]-[オプション]で設定した値を
        // 取得するためのオブジェクト
        GlobalPreferences preferences = GlobalPreferences.sharedInstance();

        // 値を更新する周期の設定
        pollingTime = preferences.getMonitoredDataPoll();
        // 値を保持しておく時間 (分)
        int cacheMin = preferences.getMonitoredDataCache();
        // 値を保持しておくサイズ
        int chartCache = cacheMin * 60 / pollingTime;

        // 初期状態でのy軸の高さ
        long yMargin = mxBean.getUsage().getCommitted();

        // y軸の単位がByteなので、
        // byteメソッドでSimpleXYChartDescriptorオブジェクトを作成
        SimpleXYChartDescriptor chartDescriptor =
                    SimpleXYChartDescriptor.bytes(yMargin, true, chartCache);

        // チャートに描画する項目
        chartDescriptor.addLineFillItems("Committed", "Used");
        
        // チャート上部に表示する項目名
        chartDescriptor.setDetailsItems(new String[] {"Max", "Committed", "Used"});

        // SimpleXYChartSupportオブジェクトの生成
        SimpleXYChartSupport chartSupport
                = ChartFactory.createSimpleXYChart(chartDescriptor);

        // チャートをパネルに貼る
        JComponent chartPanel = new JPanel();
        chartPanel.setLayout(new BorderLayout());
        chartPanel.add(chartSupport.getChart(), BorderLayout.CENTER);

        // パネルをサブタブに登録
        dataViewComponent.addDetailsView(
                new DataViewComponent.DetailsView(name, null, 0, chartPanel, null),
                position);

        return chartSupport;
    }
}
